require 'date'
require 'wise'
require 'wise/select'

Months = [
  "Jan", "Feb", "Mar", "Apr", "May", "Jun",
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
]

class Date
  def to_text
    "#{year} #{Months[month-1]} #{day}"
  end
end

module Wise
  class DateInput < Button
    # As in Select, popup needs info like Menu uses, factor it out...
    def initialize(parent, settings={})
      d = settings.delete(:date) || Date.today
      super(parent, {:text=>"0000 May 00"}.update(settings)) {
	unless @input
	  @input = VList.new(nil, :override=>true, :map=>false, :border=>1, :relief=>:up)
	  @input.extend Popup
	  @bar = HList.new(@input, :border=>0)

	  # Month manipulation
	  @month_down = Button.new(@bar, :text=>"<", :border=>0) {
	    self.date = @date << 1
	  }
	  @month = Select.new(@bar, :values=>Months, :index=>@date.month-1)
	  @month.bind("Select") { |evt|
	    self.date = Date.new(@date.year, evt.index+1, @date.day)
	  }
	  @month_up = Button.new(@bar, :text=>">", :border=>0) {
	    self.date = @date >> 1
	  }
	  Label.new(@bar, :text=>" ")

	  # Integer(!) Input for years
	  @year_down = Button.new(@bar, :text=>"<", :border=>0) {
	    set_possibly_leap_date(@date.year-1, @date.month, @date.day)
	  }
	  @year = Input.new(@bar, :text=>@date.year.to_s)
	  @year.bind("InputSet") { |evt|
	    set_possibly_leap_date(evt.data.to_i, @date.month, @date.day)
	  }
	  @year_up = Button.new(@bar, :text=>">", :border=>0) {
	    set_possibly_leap_date(@date.year+1, @date.month, @date.day)
	  }

	  # block w/ 1..31 for days
	  @ary, t = create_date_text(@date)
	  @multi = Multi.new(@input, :text=>t, :border=>0, :space=>0)
	  @multi.bind("Button") { |evt|
	    @input.unmap
	    index = evt.x / (@multi.width / 7) +
	      evt.y / (@multi.height / (@ary.length/7)) * 7
	    self.date = @ary[index]
	  }
	end
	@input.popup(root_x, root_y+self.height)
      }
      self.date = d
    end
    def pack(w, h)
      resize(w, h)  unless toplevel?
    end
    def destroy()
      @input.destroy  if @input  # should be toplevel child !!!
      super()
    end

    def set_possibly_leap_date(year, month, day)
      begin
	self.date = Date.new(year, month, day)
      rescue
	self.date = Date.new(year, month, day-1)
      end
    end
    def date()
      @date.clone
    end
    def date=(new_date)
      return  if @date == new_date.jd
      @date = new_date
      self.text = @date.to_text
      @ary, t = create_date_text(@date)
      if @input
	@month.index = @date.month - 1
	@year.text = @date.year.to_s
	@multi.text = t
	@input.dirty_height = @multi.dirty_height = true  # Wise inaccuracy !!!
	@input.resize(@input.total_width, @input.total_height)  # X dep !!!
      end
      throw_event(InputSetEvent.new(self, @date.clone))
    end

    def create_date_text(new_date)
      a = []
      day = new_date - (new_date.day-1)
      first = day - day.wday
      if first != day
	first.upto(day - 1) { |d|
	  a << d
	}
      end
      while day.month == new_date.month
	a << day
	day += 1
      end
      while a.length % 7 != 0
	a << day
	day += 1
      end
      t = ""
      a.each_with_index { |val, i|
	t += "#{'%2d'%val.day}#{i%7==6?"\n":" "}"
      }
      [a, t]
    end
    private :create_date_text
  end

  if $0 == __FILE__
    w = Window.new(nil)
    DateInput.new(w, :date=>Date.new(1998, 12, 23))
    main_thread()
  end
end
