require 'wise/native'

require 'kero/serialize'

module Wise
  DefaultDisplay = ENV["DISPLAY"] || ":0.0"
  class Display
    # Factory
    attr_accessor :root
    private :new

    @@hash = {}
    def Display::get(name=DefaultDisplay)
      return name  if name.instance_of? Display
      unless disp = @@hash.fetch(name, nil)
	p disp
	p name
	disp = Display.new(name)
	disp.root = Window.root_for(disp)
	@@hash.store(name, disp)
	disp.toplevels = []
	disp.orphans = []
      end
      disp
    end

    attr_writer :toplevels, :orphans
    def add_toplevel(window)
      #puts "@toplevels was #{@toplevels}, add #{window}"
      @toplevels << window
      if window.orphan?
	if @orphans.empty?
	  @disp_io = IO.new(fd, "r")
	  Serialize.set_input(@disp_io) {
	    nextEvent()  # handles as many events as it can get
	  }
	end
	@orphans << window
      end
    end

    def delete_toplevel(window)
      @toplevels.delete(window)
      if window.orphan?
	@orphans.delete(window)
	#puts "@orphans becomes #{@toplevels}"
	if @orphans.empty?
	  Serialize.unset_input(@disp_io)
	  # puts "We're done with Wise"
	end
      end
    end

    def draw_clean()
      @toplevels.each() { |window|
	window.draw()
      }
      flush()
    end

    def clean_layout()
      @toplevels.each() { |window|
	puts "Cleaning layout for #{window}"  if window.layout_clean?
	window.pack(window.total_width, window.total_height)  if window.layout_clean?
      }
      flush()
    end

    def handle_event(window, full_name, event)
      #puts " ** Callback for window #{window}, event #{event.serial} #{full_name}"
      return  unless window  # *argh*, get id 0 on iPAQ when rotating screen.
      ancestor = window
      if event != nil and ancestor != event.window
	puts "Ruby Wise::Display::handle_event() -- Wrong window for event!"
	puts " (probably got destroyed before the event was handled)"
      end
      names = full_name.split(/-/).reverse
      (names.length-2).downto(0) { |i|
	names[i] = names[i+1] + "-" + names[i]
      }
      while ancestor
	names.each { |name|
	  if callback = ancestor.bindings.fetch(name, nil)
	    callback.each { |cb|
	      comply = true
	      cb[0].each { |mod|
		case mod
		when "Press"
		  comply &&= event.press?
		when "Release"
		  comply &&= event.release?
		else
		  # assume detail
		end
	      }
	      if comply
		#puts "callback #{name} for #{ancestor}, mods #{cb[0]}"
		cb[1].call(event)  unless event.skip_internals
	      end
	    }
	    break  unless full_name == "Key" or full_name == "Button"
	  end
	}
	break  if ancestor.toplevel?
	break  if event.skip_parents
	# event.window can have been destroyed; due to XFlush, no race
	# conditions. But, using ancestor is dangerous (though ancestor.parent
	# may still exist)
	if full_name == "Button" and event.window and event.window.parent
	  # adapt *click* so it is in the correct place in the parent
	  # Should I also do this with a Key Event?
	  event.x += event.window.x
	  event.y += event.window.y
	  # event.window = event.window.parent
	  event.no_longer_original  # ...
	end
	ancestor = ancestor.parent
      end
      if event.kind_of? FakeEvent
	# take over from xdisplay.c:nextEvent()
	draw_clean()
	clean_layout()
      end
    end
  end
end
