require 'wise/focus'

module Wise
  class Input < Label
    include Focus
    attr_reader :text
    def initialize(parent, settings={})
      # bw = parent.get_setting(:border)
      # bw = 2  unless bw
      super(parent, {:relief=>:down, :border=>2, :align=>:left}.update(settings))
      @index = @text.length
      @old_text = @text.clone()
      bind("Key") { |evt|
	# add editing commands
	if evt.key_string
	  @text[@index, 0] = evt.key_string
	  @index += 1
	  #self.text = @text
	elsif evt.key_sym == Enter
	  maybe_throw()
	elsif evt.key_sym == Home
	  @index = 0
	elsif evt.key_sym == LeftArrow
	  @index -= 1  if @index > 0
	elsif evt.key_sym == RightArrow
	  @index += 1  if @index < @text.length
	elsif evt.key_sym == End
	  @index = @text.length
	elsif evt.key_sym == BackSpace
	  if @index > 0
	    @text[@index-1, 1] = ""
	    @index -= 1
	    #self.text = @text
	  end
	elsif evt.key_sym == Delete
	  if @index < @text.length
	    @text[@index, 1] = ""
	    #self.text = @text
	  end
	else
	  p evt.key_sym
	end
	@clean = false
	clear; draw # ; display.flush()
      }
      bind("Focus") { |evt|
	maybe_throw()  if evt.out?
      }
      bind("Button-Press") {
	take_input_focus
	toplevel.focus_sibling = self
      }
      bind("Button-Release") { |evt|
	if @down and not @moved and
	    (Time.now.to_f - @down > 0.25 or evt.detail == 2)
	  request_selection()
	else
	  min, max = 0, @text.length
	  while max - 1 > min
	    mid = (min + max) / 2
	    if @xoffset + font.width(@text[0...mid]) < evt.x
	      min = mid
	    else
	      max = mid
	    end
	  end
	  ch_lo, ch_hi = font.width(@text[0...min]), font.width(@text[0...max])
	  new_index = (@xoffset + (ch_lo+ch_hi) / 2 < evt.x) ? max : min
	  if @index != new_index
	    @index = new_index
	    @clean = false
	    clear(); draw()
	  end
	end
      }
      bind("SelectNotify") {
	sel = read_selection
	@text[@index, 0] = sel
	self.text = @text
	@index += sel.length
      }
    end

    def text=(new_text)
      @index = new_text.length  if @index > new_text.length
      @old_text = new_text.clone()
      super(new_text)
    end

    def maybe_throw()
      if @text != @old_text
	throw_event(InputSetEvent.new(self, @text))
	@old_text = @text.clone()  # wrong for multi, need deep_clone
      else
	puts "Didn't throw InputSet, 'coz content not changed"  if $DEBUG
      end
    end

    def user_width()
      if chars = settings[:chars]
	font.width("v"*chars)
      else
	super()
      end
    end

    def hoffset(to_display=font.width(@text))
      @xoffset = 0  unless @xoffset
      if to_display > width - 2 * (border + space)
	cursor_offset = @xoffset + font.width(@text[0...@index])
	calib_width = [font.width("vvv"), width / 2 - border - space].min
	if cursor_offset < border + space + calib_width
	  @xoffset += border + space + calib_width - cursor_offset
	  cursor_offset = @xoffset + font.width(@text[0...@index])
	  if cursor_offset > width / 2
	    @xoffset -= cursor_offset - width / 2
	  end
	  @xoffset = border + space  if @xoffset > border + space
	elsif cursor_offset > width - border - space - calib_width
	  @xoffset -= cursor_offset - width + border + space + calib_width
	  @xoffset = width - to_display - border - space  if @index == @text.length
	  cursor_offset = @xoffset + font.width(@text[0...@index])
	  if cursor_offset < (width+1) / 2
	    @xoffset += (width+1) / 2 - cursor_offset
	  end
	  if @xoffset + to_display <= width - border - space
	    @xoffset = width - border - space - to_display
	  end
	end
      else
	@xoffset = super(to_display)
      end
      return @xoffset
    end

    def draw_content(xoffset, yoffset)
      super(xoffset, yoffset)
      if relief == :focus
	x_cursor = xoffset + font.width(text[0...@index]) - 1
	draw_line(x_cursor, yoffset, x_cursor, yoffset + font.height)
      end
    end
  end

  class StringEdit < Input
    def initialize(parent, settings={})
      @string = settings.delete(:string)
      super(parent, {:text=>@string}.update(settings))
      bind("InputSet") { |evt|
	@string[0..-1] = evt.data
      }
    end
    def string=(new_string)
      @string = new_string  # reference
      self.text = @string  # to display
    end
  end
end
