require 'wise/window'

module Wise
  class Label < Window
    def initialize(*settings)
      parent = nil
      if not settings[-1].kind_of? Hash
	settings << {}
      end
      settings.each { |s|
	if s == nil
	  next
	elsif s.kind_of? Window
	  parent = s
	elsif s.kind_of? Hash
	  parent = s.delete(:parent)  unless parent
	  @text = s[:text]  unless @text
	  s.delete(:text)
	  # @align = s.fetch(:align, nil)
	elsif (s.instance_of? ::String)
	  @text = s
	else
	  p "Ignoring `#{s}':#{s.class}..."
	end
      }
      @text = ""  unless @text
      super(parent, {:relief=>:none}.update(settings[-1]))
      bind("Button-Press") { |evt|
	@start_index = get_index(evt.x, evt.y)
	@down = Time.now.to_f
	@moved = false
      }
      bind("Motion") { |evt|
	if @down and Time.now.to_f - @down > 0.1
	  @moved = true
	  si, ei = [@start_index, get_index(evt.x, evt.y)].sort
	  draw_selection(si, ei)
	end
      }
      bind("Button-Release") { |evt|
	if @down and @moved and Time.now.to_f - @down > 0.1
	  @start_index, @end_index = [get_index(evt.x, evt.y), @start_index].sort
	  p "#{@text[@start_index..@end_index]}"  if $DEBUG
	  # display.copy(@text[@start_index..@end_index])
	  set_selection(grab_selection)
	end
	# unbind("Motion", @proc)
	@down = false
      }
    end

    def draw_selection(si, ei)
      @clean = false
      clear; draw  # inefficient...
      draw_string(@text[si..ei],
		  hoffset + font.width(@text[0...si]),
		  @yoffset + font.ascent,
		  :invert, :dont_flush)
    end

    def grab_selection
      @text[@start_index..@end_index]
    end

    def deselect()
      # unmark the highlighted selection
      @clean = false
      clear; draw  # inefficient
    end

    def get_index(bx, by, line=@text)
      bx -= hoffset(font.width(line))
      lo, hi = 0, line.length
      while hi-lo > 1
	mid = (lo+hi) / 2
	# p "#{lo}-#{mid}-#{hi}; #{line[0, mid]} is #{font.width(line[0, mid])} wide."
	if bx <= font.width(line[0, mid])
	  hi = mid
	else
	  lo = mid
	end
      end
      return lo
    end

    def text()
      @text
    end

    def text=(new_text)
      new_text = ""  unless new_text
      @text = new_text
      @clean = false
      self.layout_dirty!
      clear; draw; display.flush()
      new_text
    end

    # The stuff with ascent and offsets is not very nice. However, text is
    # drawn from its baseline. That's good, except Wise has no concept like
    # that implemented. If it were to, user_heigth probably should return two
    # numbers: above and below baseline.
    # The alignment would change from that, since the baseline would then have
    # to be aligned. :top and :bottom can be done, but :center would be
    # max over above + max over below as height, instead of max over height as
    # is done now.
    # I don't think this would extend to a vertical baseline.

    def draw_content(xoffset, yoffset)
      @yoffset = yoffset
      draw_string(@text, xoffset, yoffset+font.ascent, :dont_flush)
      super(xoffset, yoffset)
    end

    def user_width()
      font.width(@text)
    end

    def user_height()
      font.height()
    end
  end
end
