require 'wise/list'
require 'wise/popup'

module Wise
  class Menu < VList
    include Popup

    def initialize(parent, settings={})
      super(parent, {:override=>true, :relief=>:up}.update(settings))
    end

    alias window_unmap unmap
    def unmap()
      children.each { |child| child.menu.unmap  if child.kind_of? MenuButton}
      window_unmap
    end

    attr_accessor :direction
  end

  class MenuButton < Button
    attr_reader :triangle
    def initialize(parent, settings={})
      @triangle = settings.delete(:triangle)
      @triangle = true  if @triangle == nil and parent.kind_of? Menu
      super(parent, {:space=>0, :align=>:left, :relief=>:none}.update(settings))
      @menu = Menu.new(self, settings.dup.update(:map=>false, :override=>true))
      bind("Clicked") { |evt|
	if parent.kind_of? Menu
	  parent.children.each { |child|
	    next  if child == self or not child.kind_of? MenuButton
	    child.menu.unmap
	  }
	  # [@menu.total_width, display.root.width].min
	  # [@menu.total_height, display.root.height].min
	  @menu.resize(@menu.total_width, @menu.total_height)

	  y = nil
	  if parent.direction[0] == :up
	    if root_y + height - @menu.height >= 0
	      y = root_y + height - @menu.height
	      @menu.direction = [:up, parent.direction[1]]
	    else
	      if root_y + @menu.height < display.root.height
		y = root_y
	      else
		y = 0
	      end
	      @menu.direction = [:down, parent.direction[1]]
	    end
	  else  # mostly :down
	    if root_y + @menu.height < display.root.height
	      y = root_y
	      @menu.direction = [:down, parent.direction[1]]
	    else
	      if root_y + height - @menu.height >= 0
		y = root_y + height - @menu.height
	      else
		y = display.root.height - @menu.height
	      end
	      @menu.direction = [:up, parent.direction[1]]
	    end
	  end
	    
	  if root_x+width + @menu.width < display.root.width
	    @menu.popup(root_x+width, y)
	  elsif @menu.width < root_x
	    @menu.popup(root_x - @menu.width, y)
	  else
	    @menu.popup(0, y)
	  end
 	else
	  # top of the menu, perhaps in a menubar
 	  if @menu.mapped?
 	    @menu.unmap
 	  else
 	    @menu.resize(@menu.total_width, @menu.total_height)
	    p display
	    p display.root
	    p display.root.height
 	    if root_y+height + @menu.height < display.root.height
 	      @menu.popup(root_x, root_y+height)
	      @menu.direction = [:down, :right]
 	    elsif @menu.height < root_y
 	      @menu.popup(root_x, root_y - @menu.height)
	      @menu.direction = [:up, :right]
 	    else
 	      # restrict size of VList when it supports scrollbars
 	      # for now, give the best shot you can give:
 	      @menu.popup(root_x, 0)
 	    end
 	  end
	end
      }
    end
    attr_reader :menu

    def user_width()
      super + (triangle ? user_height : 0)
    end

    def draw_content(xoffset, yoffset)
      super(xoffset, yoffset)
      if triangle
	fg_used = bg.dark
	edge = width - space - border - user_height + 1
	base = user_height - 2
	draw_line(edge, 3, edge, base)
	draw_line(edge, 3, edge+base-1, (user_height+1)/2)
	draw_line(edge, base, edge+base-1, (user_height+1)/2)
	fg_used = fg
      end
    end

    #def unmap()
    #  menu.unmap
    #  super()
    #end
  end
end
