module Wise
  module TabbedLayout
    def user_width
      (@children.map { |c| c.total_width } << @top.total_width).max
    end

    def user_height
      (@children.map { |c| c.total_height } << 0).max + @top.total_height
    end

    def pack(w, h)
      resize(w, h)  unless toplevel?
      offset = 0
      if not @top.children.empty?
	offset = @top.total_height
	@top.pack(w, offset)
      end
      @children.each { |c|
	if c == @chosen
	  c.map
	  if c.width != w or c.height != h-offset
	    c.move(0, offset)
	    c.pack(w, h-offset)
	  end
	else
	  c.unmap
	end
      }
    end
  end

  class TabbedButton < Button
    def initialize(parent, settings={})
      @ref = settings[:text]
      super(parent, {:border=>2, :layout=>NoLayout, :align=>:left}.update(settings))
      @b = Button.new(self, {:border=>0, :text=>'X', :align=>:center}) { |evt|
	parent.parent.delete_tab(@ref)
	evt.no_internals!
      }
#      @b.bind("Button-Release") { |evt|
#	evt.no_internals!
#      }
    end
    def user_width()
      super + user_height
    end
    def pack(w, h)
      if @b
	@b.move(w-border-space-user_height, border+space)
	@b.pack(user_height, user_height)
      end
    end
    def draw_content(xoffset, yoffset)
      super(xoffset, yoffset)  # no extra movement, since they're widgets
      @b.draw(0, 0)
    end
  end

  class Tabbed < Window
    def initialize(parent, settings={})
      @del_callback = settings[:del_callback]
      super(parent, {:layout=>TabbedLayout}.update(settings))
      @top = HList.new(self)
      @init_done = true
    end

    def insert_child(child, index=-1)
      return  unless @init_done
      super
      @chosen = child  unless @chosen
      tb = TabbedButton.new(@top, :text=>child.title, :assoc=>child) {
	@chosen = tb.settings[:assoc] # @children[@top.children.index(tb)]
# p @top.children.index(tb)
# puts @children
# p @chosen
	pack(total_width, total_height)
      }
    end

    def delete_tab(name)
      i = 0
      while @top.children[i].text != name
	i += 1
      end
      @top.delete_child(@top.children[i])
      tab = children[i]
      if tab == @chosen
	@chosen = children[i+2]
	@chosen = children[0]  unless @chosen
      end
      delete_child(tab)
      @del_callback.call(name)  if @del_callback
    end

    def empty
      super
      @top.empty!
      @chosen = nil
    end

    def empty!
      super
      @top.empty!
      @chosen = nil
    end

    def draw_content(xoffset, yoffset)
      @top.draw()
      super(xoffset, yoffset)
    end
  end
end
