require 'wise/native'

module Wise
  RecursiveProperties = [
    :display, :fg, :bg, :fg_used, :bg_used, :font,
    :fixed_font, :border, :space
  ]
  # :border is the odd one out, since it only adds to :space for many widgets
  # perhaps add :menu_border, since 2 is too much for a menu entry?
  # wow, perhaps I need to types of menu buttons, even...
  SingleProperties = [:native, :relief, :title, :layout]
  # isasmuch as :native is a "property"
  ToplevelProperties = [:dock, :modal, :grab, :override]
  # where does :map fit in?
  # :index fulfills an inter-widget relation, before layout managers show up
  LayoutMgrPoperties = [:side, :align, :space]
  # :space is recursive; all layout managers should use it.

  class Window
    # workaround to bootstrap the root of the display
    # it might be cleaner to construct the root only when requested
    undef_method(:display)
    def initialize(*settings)
      @settings = settings[-1]
      @native = @settings[:native]
      @width, @height = native_width(), native_height()
    end
    attr_reader :native

    def get_setting(symbol)
      if symbol == :display
	@settings[:display]
      end
    end

    def Window::root_for(display)
      p display.native_root
      return Window.new(nil, :display=>display, :native=>display.native_root)
    end
  end
  #class Input < Window
  #end
end

require 'wise/display'

module Wise
  class Color
    def light()
      @light = Color.new((red+256)/2, (green+256)/2, (blue+256)/2)  unless @light
      @light
    end
    def dark()
      @dark = Color.new(red/2, green/2, blue/2)  unless @dark
      @dark
    end
    def to_s()
      "(#{red}, #{green}, #{blue})"
    end

    Black = Color.new(0, 0, 0)
    Green = Color.new(100, 250, 100)
    Red = Color.new(250, 100, 100)
    Blue = Color.new(100, 100, 250)
    White = Color.new(255, 255, 255)
    PostIt = Color.new(255,250,190)
  end
end

require 'wise/event'
require 'wise/layout'
require 'wise/scrollbar'

module Wise

  DefaultRecursiveProperties = {
    :bg=>Color.new(200, 210, 220), :fg=>Color.new(20, 60, 110),
    :bg_used=>Color.new(200, 210, 220), :fg_used=>Color.new(20, 60, 110),
    :font=>Font.new("lucida"), :fixed_font=>Font.new("fixed"),
    :border=>2, :space=>1
  }

  class Window
    attr_reader :parent, :settings
    def initialize(*sett)
      sett.shift  while not sett[0] and not sett.empty?
      if sett[0] and not sett[0].kind_of? Hash
	@parent = sett.shift
      else
	@parent = sett[:parent]  # could be nil
      end
      while sett[0] and not sett[0].kind_of? Hash
	p sett[0]
	option = sett.shift
	p "Dropping #{option}..."
      end
      sett = sett[0] || {}
      @settings = sett  # Supposedly Hash...

      # if grabbing root of display
      if @settings[:native]
	p "native found (root window)"
	@native = @settings[:native]
	return
      end

      # fixing (extra) settings
      @settings[:bg_used] = @settings[:bg]  if @settings[:bg]
      @settings[:fg_used] = @settings[:fg]  if @settings[:fg]
      if @settings[:font] and not @settings[:font].instance_of? Font
	@settings[:font] = Font.new(@settings[:font])
      end

      if orphan?
	# toplevel and lookup for descendants
	disp = Display.get(@settings.fetch(:display, DefaultDisplay))
	@settings.store(:display, disp)  # replace String by Wise::Display
      else
	@settings[:display] = parent.display
      end
      if toplevel?
	display.add_toplevel(self)
	@settings = {:map=>true, :title=>"Wise"}.update(@settings)
      end

      @children = []
      @toplevels = []

      @layout_set = false
      self.layout_clean = false

      @width, @height = total_width, total_height
      @old_width, @old_height = 0, 0  if toplevel?
      @native = create_native(parent, self)
      if parent
	if toplevel?
	  parent.add_toplevel(self)
	else
	  parent.insert_child(self, @settings.fetch(:index, nil))
	end
      end

      @bindings = {}
      bind("Expose") { make_dirty }
      bind("ConfigureNotify") { |evt|
	#puts "ConfigureNotify for #{evt.window}"
	@width, @height = native_width, native_height
	if evt.window == self and toplevel? and mapped?
	  #puts "Configure Notify #{evt.serial} on a mapped toplevel window..."
	  if width != @old_width or height != @old_height
	    #puts "...repacking to #{width}x#{height}"
	    pack(width, height)
	    @old_width, @old_height = width, height
	    # sleep 0.2
	  end
	end
      }
      unless parent
	bind("Focus") { |evt|
	  #puts "Focus win=#{evt.window}, in?#{evt.in?}, detail:#{evt.detail}"
	  if evt.in? and evt.window == self and evt.detail == NotifyNonlinear
	    if focus_sibling
	      focus_sibling.take_input_focus
	    else
	      fs = first_sibling
	      focus = fs
	      while focus and (not focus.respond_to?("take_input_focus") or not focus.mapped?)
		focus = focus.next_sibling
	      end
	      focus.take_input_focus  if focus
	    end
	  elsif evt.out?
	    if focus_sibling
	      focus_sibling.lost_input_focus
	    end
	  end
	}
      end
      if @settings[:tooltip]
	tooltip = Label.new(self, :override=>true, :text=>@settings[:tooltip],
			    :map=>false, :bg=>Color::PostIt, :fg=>Color::Black)
	tooltip.extend Popup
	@prev_pop_time = Time.now.to_f - 2.00
	bind("Enter") { |evt|
	  if Time.now.to_f - @prev_pop_time >= 2.00
	    @pop_thread = Thread.new {
	      sleep(0.25);
	      puts "#{evt.x} + #{root_x}, #{evt.y} + #{root_y}"  if $DEBUG
	      tooltip.popup(evt.x + root_x, evt.y + root_y)
	      sleep(1.75);
	      tooltip.unmap()
	    }
	  end
	  @prev_pop_time = Time.now.to_f
	}
	# bind("Motion") { @pop_thread.kill  if @pop_thread; tooltip.unmap }
	tooltip.bind("Motion") { @pop_thread.kill; @pop_thread = nil; tooltip.unmap() }
      end
      bind("Destroy") { |evt|
	destroy  if evt.window == self
      }
      @map = @settings.fetch(:map, true)  if toplevel?
      @settings.delete(:map)
      make_dirty
      # throw_event(FakeEvent.new(self, "Expose"))  if mapped?
    end
    attr_reader :children, :native, :bindings
    attr_accessor :focus_sibling
    attr_accessor :width, :height

    # Wise toplevel, but not modal, overridden or docked.
    # used as top for inheriting RecursiveProperties
    def orphan?
      @parent == nil
    end

    # Wise toplevel, there can always be a WM or else the (X) root window
    def toplevel?()
      orphan? or @settings[:modal] or @settings[:override] or @settings[:dock]
    end

    def toplevel()
      ancestor = self
      while not ancestor.toplevel?
	ancestor = ancestor.parent
      end
      ancestor
    end

    attr_writer :clean
    def clean?()  @clean  end
    def layout_clean?()  @layout_clean  end
    def layout_clean=(val)
      @layout_clean = val
      @cache_width = @cache_height = nil  unless val
    end

    def make_dirty()  # exposed/image/display/whatever dirty
      ancestor = self
      begin
	ancestor.clean = false
	ancestor = ancestor.parent
      end  while ancestor and ancestor.clean?
    end

    def layout_dirty!()
      self.layout_clean = false
      ancestor = self
      while not ancestor.toplevel? # and ancestor.parent.layout_clean?
	ancestor = ancestor.parent
	ancestor.layout_clean = false
      end
    end

    def display
      get_setting(:display)
    end

    def display=(display)
      raise "Can't set Display!"
    end

    def title
      get_setting(:title)
    end

    def add_toplevel(child)
      @toplevels << child
    end

    def insert_child(child, index=nil)
      if child.parent and child.parent != self
	# reparent in X
	child.parent.children.delete(child)
      end
      if not @layout_set
	extend settings.fetch(:layout, Layout)
	@layout_set = true
      end
      if index
	@children[@children.empty? ? 0 : index, 0] = child
      else
	@children << child
      end
      pa = self
      begin
	pa.layout_dirty!
	pa = pa.parent
      end while pa
      if toplevel.mapped?
	toplevel.resize(toplevel.total_width, toplevel.total_height)
	# toplevel.pack(toplevel.width, toplevel.height)  # X dep !!!
      end
    end

    def delete_child(child)
      @children.delete(child)
      child.unmap
      ancestor = self
      begin
	ancestor.layout_dirty!
	ancestor = ancestor.parent
      end while ancestor
      if toplevel.mapped?
	toplevel.resize(toplevel.total_width, toplevel.total_height)
	#toplevel.pack(toplevel.width, toplevel.height)
      end
    end

    # simply throw all children out ...
    def empty
      @children = []
    end

    # ... but most of the time you'll want auto-destroy
    def empty!
      @children.each { |c|
	c.destroy
      }
      @children = []
    end

    def first_sibling()
      ancestor = self
      while ancestor.children[0]
	ancestor = ancestor.children[0]
      end
      ancestor
    end

    def next_sibling()
      prev = self
      prt = prev.parent
      while prt != nil and (index = prt.children.index(prev)) != nil and prt.children[index+1] == nil
	prev = prt
	prt = prev.parent
      end
      return nil  if prt == nil or index == nil
      prt.children[index+1].first_sibling
    end

    # called from C-side
    def map=(new_status)
      @map = new_status
    end
    private :map=

    def mapped?
      @map
    end

    alias native_unmap unmap
    def unmap()
      @toplevels.each { |child| child.unmap }
      native_unmap()
    end

    def halign()
      result = @settings[:halign]
      return result  if result
      align = @settings[:align]
      return align  if align == :left or align == :right
      return :center
    end

    def valign()
      result = @settings[:valign]
      return result  if result
      align = @settings[:align]
      return align  if align == :top or align == :bottom
      return :center
    end

    def hoffset(to_display=user_width)
      if halign == :left
	border + space
      elsif halign == :right
	width - border - space - to_display
      else
	(width - to_display) / 2
      end
    end

    def side()
      @settings.fetch(:side, :top)
    end

    def voffset(to_display=user_height)
#     @vscroll = Scrollbar.new(parent)  unless @vscroll
      if to_display > height and @vscroll
	@vscroll.partially(width, to_display)
	@vscroll.view[0]
      else 
	if valign == :top
	  border + space
	elsif valign == :bottom
	  height - border - space - to_display
	else
	  (height - to_display) / 2
	end
      end
    end

    def pack(w, h)
      resize(w, h)  unless toplevel?
    end

    def layout=(new_layout)
      @settings[:layout] = new_layout
      @layout_set = false
    end

    def draw(xoffset=0, yoffset=0, cached_settings={})
      return  if @clean
      @cached_settings = cached_settings.dup.update(@settings)
      # compute offsets, based on alignment, border and (when av.) scrollbars
      draw_content(xoffset+hoffset, yoffset+voffset)
      b = border
      r = relief
      draw_border(b)  if b != 0 and r and r != :none
      @clean = true
      @cached_settings = nil
    end

    def draw_content(xoffset, yoffset)
      # childen WIDGETS have a placement, hence don't need an offset
      children.each { |ch| ch.draw(0, 0, @cached_settings) }
      @toplevels.each { |ch| ch.draw(0, 0, @cached_settings) }
    end
    alias draw_content_Window draw_content

    def draw_border(border_width=border)
      # return  unless border_width
      r = relief
      # return  if not r or r == :none  # or bg == Transparent
      w, h = width, height
      case r
      when :focus
	self.fg_used = Color::Black
      when :up
	self.fg_used = bg.light
      else
	self.fg_used = (toplevel? ? bg.dark : parent.bg.dark)
      end
      border_width.times { |i|
	draw_line(i, i, w-2-i, i, :dont_flush)
	draw_line(i, i+1, i, h-2-i, :dont_flush)
      }
      case r
      when :focus
	self.fg_used = Color::Black
      when :up
	self.fg_used = bg.dark
      else
	self.fg_used = (toplevel? ? bg.light : parent.bg.light)
      end
      border_width.times { |i|
	draw_line(i+1, h-1-i, w-1-i, h-1-i, :dont_flush)
	draw_line(w-1-i, i+1, w-1-i, h-2-i, :dont_flush)
      }
      self.fg_used = fg
    end

    def relief()  get_setting(:relief, :none)  end

    def relief=(new_relief)
      @settings.store(:relief, new_relief)
      @clean = false
      draw
    end

    # User width and height, are typically the result of what is displayed
    # inside the widget. Override as often as possible in subclasses (new
    # widgets, as well as layout managers for composed widgets).
    def user_width()
      @settings.fetch(:width, 100)
    end
    def user_height()
      @settings.fetch(:height, 100)
    end

    def border()
      relief = @settings[:relief] || :none
      if relief == :up or relief == :down or relief == :focus
	return get_setting(:border, 0)
      end
      # should be using space when there is no relief, anyway
      return 0
    end
    def space()  get_setting(:space, 0)  end

    # Distinguish between left/right space & border, as well as top/bottom?
    # I know kaffe does it and it seems to make sense for top-level windows
    # (which borders we'll need to implement for WM-less environments, once).
    def total_width()
      return @cache_width  if @cache_width
      result = user_width
      result += 2*(space + border)
      result = 1  if result == 0  # XCreateWindow workaround
      @cache_width = result
    end

    def total_height()
      return @cache_height  if @cache_height
      result = user_height
      result += 2*(space + border)
      result = 1  if result == 0  # XCreateWindow workaround
      @cache_height = result
    end

    def bg() get_setting(:bg) end
    def bg_used() get_setting(:bg_used) end
    def fg() get_setting(:fg) end
    def fg_used() get_setting(:fg_used) end
    def font() get_setting(:font) end
    def fixed_font() get_setting(:fixed_font) end

    def bg=(new_bg)
      @settings[:bg] = new_bg
      @settings[:bg_used] = new_bg
      clear; draw
    end

    def bg_used=(new_used)
      @settings[:bg_used] = new_used
      @clean = false
      clear; draw
    end
    protected :bg_used=

    def fg=(new_fg)
      @settings[:fg] = new_fg
      @settings[:fg_used] = new_fg
      # clear; draw
    end

    def fg_used=(new_used)
      @settings[:fg_used] = new_used
    end
    protected :fg_used=

    def get_setting(symbol, default=nil)
      #if symbol.class != Symbol
      #	raise "#{symbol} of class #{symbol.class}, not Symbol"
      #end
      #caller.each { |c| p c }  unless @settings
      result = @settings[symbol]
      #print("+")  if result; $stdout.flush()
      return result  if result
      if RecursiveProperties.include? symbol
	if @cached_settings
	  result = @cached_settings[symbol]
	  #print("!")  if result;  $stdout.flush()
	  return result  if result
	end
	ancestor = parent
	while not result and ancestor
	  result = ancestor.settings.fetch(symbol, nil)
	  ancestor = ancestor.parent
	end
	#print("-")  if result; $stdout.flush()
	return result  if result
	#print("?")  if result; $stdout.flush()
	result = DefaultRecursiveProperties[symbol]
	return result  if result
      end
      return default
    end

    def method_missing(symbol, *arg)
      #p "method_missing(#{symbol}, #{arg})"
      name = symbol.to_s()
      if name[-1, 1] == "=" and arg.length == 1
	@settings.store(name[0..-2].intern, arg[0])
      else
	super(symbol, arg)
      end
    end


    def destroy()
      display.delete_toplevel(self)  if orphan?
      @children.each { |child| child.destroy }
      @toplevels.each { |child| child.destroy }
      @bindings.clear
      destroy_native()
    end


    # A mask is Main-Detail?   (multiple details possible in the future)
    # [ => in the future, multiple Mods can be prefixed to this ]
    # We need to be able to find at least the Main

    Events = ["Expose", "ConfigureNotify", "Destroy",
      "Key", "Button",
      "Motion", "Focus", "Enter", "Leave",
      "SelectNotify"
    ] + ["Clicked", "Scrolled",
      "Scroll-LineUp", "Scroll-LineDown", "Scroll-PageUp", "Scroll-PageDown",
      "Select", "InputSet"
    ]

    Mains = []
    Events.each { |mask|
      mask.split(/-/).each { |submask|
	Mains << submask  unless Mains.include? submask
      }
    }

    def bind(*mask, &hekje)
      mask.each { |m|
	subs = m.split('-')
	name = nil
	subs.each { |mod|
	  if Mains.include? mod
	    name = mod
	    break
	  end
	}
	raise  "No name found in #{m} (#{subs})"  unless name
	subs.delete(name)
	if name == "Key" or name == "Button"
	  # fix default Detail for Key- and Button-Events
	  unless subs.include? "Press" or subs.include? "Release"
	    subs << "Release"
	  end
	end
	#puts "#{self} subscribes to #{name} + #{subs}"
	ary = @bindings.fetch(name, [])
	ary.unshift([subs, hekje])  unless ary.include? [subs, hekje]
	@bindings.store(name, ary)
	if (name == "Expose") and mapped?
	  #  small hack to support POLS
	  throw_event(FakeEvent.new(self, "Expose"))
	end
	display.listen_to(self, name)
      }
    end

    #def unbind(mask, &hekje)
    #  ary = @bindings.fetch(mask, [])
    #  ary.delete(hekje)
    #  # display.deaf(self, mask)  iff ary.empty? and mask.parents.empty?
    #end

    # Throw an event, synchronously.
    # Will need an asynchronous one, too (perhaps use X for it?)
    def throw_event(event)
      display.handle_event(self, event.name, event)
    end
  end

  def Wise::main_thread()
    Serialize::serialize()
  end
end
